// Copyright ©2016 The Gonum Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Package fonts provides the Liberation fonts from https://fedorahosted.org/liberation-fonts/
package fonts // import "gonum.org/v1/plot/vg/fonts"

//go:generate go run mk-fonts.go
